using System;
using GameKnight.Core;

namespace GameKnight
{
	
	namespace GameBoy
	{
	
		public partial class VideoMemory : IVideoMemory, ISerializable
		{

			private readonly AttributeTable[] AttributeTableArray = new AttributeTable[2];

			private readonly Register.BNK BNK = null;

			private readonly Map[] MapArray = new Map[2];

			private readonly NameTable[] NameTableArray = new NameTable[2];

			private readonly IPalette[] PaletteArray = null;

			private readonly PatternTable[] PatternTableArray = new PatternTable[4];

			private readonly Sprite[] SpriteArray = new Sprite[0x40];

			public VideoMemory( IConsole Console )
			{
				BNK = Console.GetRegister<Register.BNK>();
				if( BNK != null )
				{
					this.PaletteArray = new IPalette[2];
                	this.PaletteArray[0] = Console.GetRegister<Register.BCP>();
					this.PaletteArray[1] = Console.GetRegister<Register.OCP>();
				}
				else
				{
					this.PaletteArray = new IPalette[3];
                	this.PaletteArray[0] = Console.GetRegister<Register.BGP>();
					this.PaletteArray[1] = Console.GetRegister<Register.OBP0>();
					this.PaletteArray[2] = Console.GetRegister<Register.OBP1>();
				}
				this.NameTableArray[0] = new NameTable( this, 0x9800 );
				this.NameTableArray[1] = new NameTable( this, 0x9C00 );
				this.AttributeTableArray[0] = new AttributeTable( this, 0xB800 );
				this.AttributeTableArray[1] = new AttributeTable( this, 0xBC00 );
				this.PatternTableArray[0] = new PatternTable( this, 0x8000 );
				this.PatternTableArray[1] = new PatternTable( this, 0x8800 );
				this.PatternTableArray[2] = new PatternTable( this, 0xA000 );
				this.PatternTableArray[3] = new PatternTable( this, 0xA800 );
				this.MapArray[0] = new Map( Console, NameTableArray[0], AttributeTableArray[0] );
				this.MapArray[1] = new Map( Console, NameTableArray[1], AttributeTableArray[1] );
				for( int i = 0; i < SpriteArray.Length; i++ )
				{
					this.SpriteArray[i] = new Sprite( Console, this, ( 0xFE00 | i ) );
				}
				return;
			}
			
			public IAttributeTable[] GetAttributeTables()
			{
            	return AttributeTableArray;
			}
			
			public IMap[] GetMaps()
			{
            	return MapArray;
			}

			public INameTable[] GetNameTables()
			{
            	return NameTableArray;
			}
			
			public IPalette GetPalette( int Index )
			{
				if( ( Index >= 0 ) && ( Index < PaletteArray.Length ) )
				{
					return PaletteArray[Index];
				}
				return null;
			}

			public IPalette[] GetPalettes()
			{
				return PaletteArray;
			}

			public IPatternTable GetPatternTable( int Index )
			{
            	if( ( Index >= 0 ) && ( Index < PatternTableArray.Length ) )
				{
					return PatternTableArray[Index];
				}
				return null;
			}

			public IPatternTable[] GetPatternTables()
			{
            	return PatternTableArray;
			}

			public ISprite[] GetSprites()
			{
				ISprite[] NewSpriteArray = new ISprite[40];
				Array.Copy( SpriteArray, NewSpriteArray, NewSpriteArray.Length );
				return NewSpriteArray;
			}
			
			public byte ReadU8( int Address )
			{
				switch( Address & 0x3C00 )
				{
					case 0x0000:
					case 0x0400:
					case 0x0800:
					case 0x0C00:
						return PatternTableArray[0][Address];
					case 0x1000:
					case 0x1400:
						return PatternTableArray[1][Address];				
					case 0x1800:
						return NameTableArray[0][Address];
					case 0x1C00:
						return NameTableArray[1][Address];
					case 0x2000:
					case 0x2400:
					case 0x2800:
					case 0x2C00:
						return PatternTableArray[2][Address];
					case 0x3000:
					case 0x3400:
						return PatternTableArray[3][Address];				
					case 0x3800:
						return AttributeTableArray[0][Address];
					case 0x3C00:
						return AttributeTableArray[1][Address];
					default:
						break;
				}
				return 0;
			}

			public ushort ReadU16( int Address )
			{
				return (ushort)( ReadU8( Address ) | ( ReadU8( Address + 1 ) << 8 ) );
			}
			
			public void WriteU8( int Address, byte Value )
			{
				switch( Address & 0x3C00 )
				{
					case 0x0000:
					case 0x0400:
						PatternTableArray[0][Address] = Value;	
						break;
					case 0x0800:
					case 0x0C00:
						PatternTableArray[0][Address] = PatternTableArray[1][Address] = Value;	
						break;
					case 0x1000:
					case 0x1400:
						PatternTableArray[1][Address] = Value;	
						break;
					case 0x1800:
						NameTableArray[0][Address] = Value;
						break;
					case 0x1C00:
						NameTableArray[1][Address] = Value;	
						break;
					case 0x2000:
					case 0x2400:
						PatternTableArray[2][Address] = Value;	
						break;
					case 0x2800:
					case 0x2C00:
						PatternTableArray[2][Address] = PatternTableArray[3][Address] = Value;	
						break;
					case 0x3000:
					case 0x3400:
						PatternTableArray[3][Address] = Value;	
						break;
					case 0x3800:
						AttributeTableArray[0][Address] = Value;	
						break;
					case 0x3C00:
						AttributeTableArray[1][Address] = Value;	
						break;
					default:
						break;
				}
				return;
			}
			
			public void WriteU16( int Address, ushort Value ) 
			{
				WriteU8( Address, (byte)( Value & 0xFF ) );
				WriteU8( Address + 1, (byte)( Value >> 8 ) );
				return;
			}
			
		}
		
	}
	
}
